#include <time.h>
#include <windows.h>
#include <stdio.h>

#include "..\\uart16550_win.h"

static int testno=0;


#define CHECK(x) \
    do { \
        if ((x) != ERROR_SUCCESS) { \
            printf("[%s]:%s:%d: ", __FILE__, \
            __FUNCTION__, __LINE__); \
            ErrorExit(#x); \
            ExitProcess(EXIT_FAILURE);\
        } \
    } while(0)
	
	
void ErrorExit(LPTSTR lpCond) {
	// Retrieve the system error message for the last-error code

	LPVOID lpMsgBuf;
	DWORD dw = GetLastError();

	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER |
		FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		dw,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR) &lpMsgBuf,
		0, NULL );

	// Display the error message and exit the process
	fprintf_s(stderr, "[%s]\n%s\n", lpCond, lpMsgBuf);

	LocalFree(lpMsgBuf);
}

#define BUFFER_SIZE 256

int test(int t, char * s, ...)
{
	va_list va;
	int i, output_chars=0,v;
        output_chars=printf("test %04d: ", ++testno);
	va_start(va, s);
	output_chars+=vprintf(s, va);
	va_end(va);
	fflush(stdout);
	for(i=0; i<60-output_chars; i++)
		putchar('.');
	if (!(v=(t)))
	        printf("failed\n");
	else
		printf("passed\n");
	fflush(stdout);
	return v;
}

#define fatal_test(t, s, ...) if (!test(t, s, __VA_ARGS__)) abort()

int DoReadFile(HANDLE fd, char *buffer, int size)
{
	DWORD n, from = 0;

	while (1) {
		if (!ReadFile(fd, &buffer[from], size-from, &n, NULL))
			return -1;
		if (n+from == size)
			return 0;
		from+=n;
	}
}

int DoWriteFile(HANDLE fd, char *buffer, int size)
{
	DWORD n, from = 0;

	while (1) {
		if (!WriteFile(fd, &buffer[from], size-from, &n, NULL))
			return -1;
		if (n+from == size)
			return 0;
		from+=n;
	}
}

void DoWriteRegistry(DWORD option)
{
	WCHAR path[100] = L"SYSTEM\\CurrentControlSet\\Services\\hello\\Parameters";
	WCHAR parameter[100] = L"parametru";
	WCHAR date[20] = L"3133222345";
	DWORD result;
	HKEY key;
	CHECK(
		RegCreateKeyExW(HKEY_LOCAL_MACHINE,
				path, 	// lpSubKey
				0,  	// reserved 
				NULL, 	// lpClass
				REG_OPTION_VOLATILE,	//dwOptions - key will change
				KEY_ALL_ACCESS,	// samDescriptor
				NULL,	//lpSecurityAttributes
				&key,   // for close
				&result));
			
	//CbSprintf(date,"%d",option);
	//sprintf_s(date,20,"%d",option);
	CHECK(RegSetValueExW(key, 	parameter, 0, REG_DWORD, &option, sizeof(DWORD)));
	RegCloseKey(key);
	return;	
}

void test_group(const char *com1_path, const char *com2_path, int baud, int loops)
{
	HANDLE com1, com2;
	int i, j;
	struct uart16550_line_info uli = { baud, UART16550_LEN_8, UART16550_PAR_NONE, UART16550_STOP_1};
	unsigned char wb[BUFFER_SIZE], rb[BUFFER_SIZE];
	time_t start, stop;

	start=time(NULL);
	

	fatal_test((com1=CreateFile(com1_path, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING,
						FILE_ATTRIBUTE_NORMAL, NULL)) != INVALID_HANDLE_VALUE, "open %s", com1_path);
	fatal_test((com2=CreateFile(com2_path, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING,
						FILE_ATTRIBUTE_NORMAL, NULL)) != INVALID_HANDLE_VALUE, "open %s", com2_path);
	test(DeviceIoControl(com1, 0xdeadbeef, &uli, sizeof(uli), 
			NULL, 0,  &i, NULL) == 0, "dummy DeviceIoControl %s", com1_path); 
	test(DeviceIoControl(com2, 0xdeadbeef, &uli, sizeof(uli), 
			NULL, 0,  &i, NULL) == 0, "dummy DeviceIoControl %s", com2_path); 
	test(DeviceIoControl(com1, UART16550_IOCTL_SET_LINE, &uli, sizeof(uli), 
			NULL, 0,  &i, NULL) != 0, "set line %s", com1_path); 
	test(DeviceIoControl(com2, UART16550_IOCTL_SET_LINE, &uli, sizeof(uli), 
			NULL, 0,  &i, NULL) != 0, "set line %s", com2_path); 

	for(j = 0; j < loops; j++) {
		for(i = 0; i < BUFFER_SIZE; i++) {
			wb[i] = rand() % 256;
			rb[i] = 0;
		}
		test(DoWriteFile(com1, wb, BUFFER_SIZE) == 0, "write %s", com1_path);
		test(DoReadFile(com2, rb, BUFFER_SIZE) == 0, "read %s", com2_path);
		for(i = 0; i < BUFFER_SIZE; i++) {
			if (rb[i] != wb[i])
				break;
	    	}
	    	if (i >= 16)

				test(1, "data check");
			else {
				for(i = 0; i < BUFFER_SIZE; i++)
					if (rb[i] != wb[i])
						fprintf(stderr, "%d %x %x\n", i, rb[i], wb[i]);
				test(0, "data check");
		}
	}
	test(CloseHandle(com1) != 0, "close %s", com1_path);
	test(CloseHandle(com2) != 0, "close %s", com2_path);
	stop=time(NULL);

	test(stop-start <= 15, "time (%ds) <= 15s", stop-start); 
}

int main(int argc, char **argv)
{
	srand(time(NULL));
	
	
	
#define OPTION_SENDER_ONLY 2
#define OPTION_RECEIVER_ONLY 1
	printf("==== Module insertion and port&interrupt registration ============\n");
	DoWriteRegistry(OPTION_SENDER_ONLY);
	fatal_test(system("driver load objchk_wnet_x86\\i386\\uart16550.sys") == 0, "driver load objchk_wnet_x86\\i386\\uart16550.sys");
	DoWriteRegistry(OPTION_RECEIVER_ONLY);
	fatal_test(system("driver load _test\\solutie.sys") == 0, "driver load _test\\solutie.sys");	
	
	fatal_test( system("driver unload uart16550") == 0, "driver unload uart16550");
	fatal_test( system("driver unload solutie") == 0, "driver unload solutie");
	
	
	printf("==== Reading at base speed (2400 baud) ===========================\n");
	DoWriteRegistry(OPTION_RECEIVER_ONLY);
	fatal_test(system("driver load objchk_wnet_x86\\i386\\uart16550.sys") == 0, "driver load objchk_wnet_x86\\i386\\uart16550.sys");
	DoWriteRegistry(OPTION_SENDER_ONLY);
	fatal_test(system("driver load _test\\solutie.sys") == 0, "driver load _test\\solutie.sys");	
	test_group("\\\\.\\uart1", "\\\\.\\uart0", UART16550_BAUD_2400, 10);
	fatal_test( system("driver unload uart16550") == 0, "driver unload uart16550");
	fatal_test( system("driver unload solutie") == 0, "driver unload solutie");
	
	
	printf("==== Writing at base speed (2400 baud) ===========================\n");
	DoWriteRegistry(OPTION_SENDER_ONLY);
	fatal_test(system("driver load objchk_wnet_x86\\i386\\uart16550.sys") == 0, "driver load objchk_wnet_x86\\i386\\uart16550.sys");
	DoWriteRegistry(OPTION_RECEIVER_ONLY);	
	fatal_test(system("driver load _test\\solutie.sys") == 0, "driver load _test\\solutie.sys");	
	test_group("\\\\.\\uart1", "\\\\.\\uart0", UART16550_BAUD_2400, 10);
	fatal_test( system("driver unload uart16550") == 0, "driver unload uart16550");
	fatal_test( system("driver unload solutie") == 0, "driver unload solutie");
	
	
	printf("==== Reading/Writing tests at different speeds ==================\n");
	DoWriteRegistry(OPTION_RECEIVER_ONLY);	
	fatal_test(system("driver load objchk_wnet_x86\\i386\\uart16550.sys") == 0, "driver load objchk_wnet_x86\\i386\\uart16550.sys");
	DoWriteRegistry(OPTION_SENDER_ONLY);
	fatal_test(system("driver load _test\\solutie.sys") == 0, "driver load _test\\solutie.sys");	
	test_group("\\\\.\\uart1", "\\\\.\\uart0", UART16550_BAUD_9600, 20);
	test_group("\\\\.\\uart0", "\\\\.\\uart1", UART16550_BAUD_9600, 20);
	test_group("\\\\.\\uart1", "\\\\.\\uart0", UART16550_BAUD_9600, 20);
	test_group("\\\\.\\uart0", "\\\\.\\uart1", UART16550_BAUD_9600, 20);
	fatal_test( system("driver unload uart16550") == 0, "driver unload uart16550");
	fatal_test( system("driver unload solutie") == 0, "driver unload solutie");
	
	DoWriteRegistry(OPTION_RECEIVER_ONLY);	
	fatal_test(system("driver load objchk_wnet_x86\\i386\\uart16550.sys") == 0, "driver load objchk_wnet_x86\\i386\\uart16550.sys");
	DoWriteRegistry(OPTION_SENDER_ONLY);
	fatal_test(system("driver load _test\\solutie.sys") == 0, "driver load _test\\solutie.sys");	
	test_group("\\\\.\\uart1", "\\\\.\\uart0", UART16550_BAUD_115200, 20);
	test_group("\\\\.\\uart0", "\\\\.\\uart1", UART16550_BAUD_115200, 20);
	test_group("\\\\.\\uart1", "\\\\.\\uart0", UART16550_BAUD_115200, 20);
	test_group("\\\\.\\uart0", "\\\\.\\uart1", UART16550_BAUD_115200, 20);
	fatal_test( system("driver unload uart16550") == 0, "driver unload uart16550");
	fatal_test( system("driver unload solutie") == 0, "driver unload solutie");
	
	return 0;
}
